import logging
import os
import pathlib
import shirley as sh
import sys
import torch
import transformers
import uuid
from .client import Client
from collections import OrderedDict
from typing import Any, Dict, Generator, List


logger = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


class Chat(Client):

    def __init__(self, local: bool, *args, **kwargs) -> None:
        super().__init__(local=local)

        self._device: torch.device | None = torch.device('cpu')
        self._device_name: str | None = None
        self._tokenizer: transformers.PreTrainedTokenizer | None = None
        self._model: transformers.PreTrainedModel | None = None


    @property
    def model(self) -> transformers.PreTrainedModel | None:
        return self._model


    def get_models(self) -> List[str]:
        if self.local:
            models_directory = os.path.abspath(os.path.expanduser('./models'))
            pretrained_models = [
                filename for filename in os.listdir(models_directory)
                if os.path.isdir(os.path.join(models_directory, filename))
            ]
            return pretrained_models
        else:
            return ['Qwen/Qwen-VL-Chat']


    def get_model_name_or_path(self, model_name: str) -> str:
        if self.local:
            return os.path.abspath(os.path.expanduser(f'./models/{model_name}'))
        else:
            return model_name


    def get_model_config(self) -> Dict:
        model_config = self._model.config.to_dict()
        return OrderedDict([
            ('device', self._device),
            ('device_name', self._device_name),
            ('architectures', model_config['architectures']),
            ('bf16', model_config['bf16']),
            ('fp16', model_config['fp16']),
            ('fp32', model_config['fp32']),
            ('model_type', model_config['model_type']),
            ('tokenizer_type', model_config['tokenizer_type']),
            ('torch_dtype', model_config['torch_dtype']),
            ('transformers_version', model_config['transformers_version']),
        ])


    def load_model(self, pretrained_model_name_or_path: str) -> None:
        if not os.path.exists(pretrained_model_name_or_path):
            logger.warning(f'Pre-trained model not found in path \'{pretrained_model_name_or_path}\'.')
            logger.info(f'Using remote pre-trained model \'{pretrained_model_name_or_path}\' from 🤗 Hugging Face.')
            local_files_only = False
        else:
            logger.info(f'Using local pre-trained model \'{pretrained_model_name_or_path}\'.')
            local_files_only = True

        if torch.cuda.is_available():
            self._device = torch.device('cuda')
            self._device_name = torch.cuda.get_device_name(torch.cuda.current_device())
        # elif torch.backends.mps.is_available():
        #     self._device = torch.device('mps')
        else:
            self._device = torch.device('cpu')

        tokenizer: transformers.PreTrainedTokenizer = transformers.AutoTokenizer.from_pretrained(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            local_files_only=local_files_only,
            trust_remote_code=True,
        )

        model: transformers.PreTrainedModel = transformers.AutoModelForCausalLM.from_pretrained(
            pretrained_model_name_or_path=pretrained_model_name_or_path,
            local_files_only=local_files_only,
            trust_remote_code=True,
        )

        model.generation_config = transformers.GenerationConfig.from_pretrained(
            pretrained_model_name=pretrained_model_name_or_path,
            local_files_only=local_files_only,
            trust_remote_code=True,
        )

        if model.generation_config.pad_token_id is not None:
            model.generation_config.pad_token_id = torch.tensor(
                data=[model.generation_config.pad_token_id],
                device=self._device,
            )
        if model.generation_config.eos_token_id is not None:
            model.generation_config.eos_token_id = torch.tensor(
                data=[model.generation_config.eos_token_id],
                device=self._device,
            )

        self._tokenizer = tokenizer
        self._model = model.to(device=self._device)


    def chat_stream(self, query: sh.types.QwenQuery, history: sh.types.QwenHistory = None) -> Generator[str, Any, None]:
        return self._model.chat_stream(tokenizer=self._tokenizer, query=query, history=history)


    def draw_bbox_on_latest_picture(self, history: sh.types.QwenHistory) -> str | None:
        response = history[-1][1]
        image = self._tokenizer.draw_bbox_on_latest_picture(response=response, history=history)
        if image is not None:
            images_tempdir = pathlib.Path(self.tempdir) / 'images'
            images_tempdir.mkdir(exist_ok=True, parents=True)
            name = f'img-{uuid.uuid4()}.jpg'
            filename = images_tempdir / name
            image.save(str(filename))
            logger.info(f'Image file saved in {str(filename)}.')
            return str(filename)
        return None



import gradio as gr
import numpy as np
import pathlib
from gradio.components.multimodal_textbox import MultimodalValue
from typing import List, Sequence, Tuple


QwenQuery = str
QwenHistory = List[Tuple[str, str]]

GradioComponents = gr.components.Component | Sequence[gr.components.Component] | None

DropdownInput = str | int | float | List[str | int | float] | List[int | None] | None
DropdownOutput = str | int | float | List[str | int | float] | None

ChatbotTuplesInput = List[List[str | Tuple[str, str] | gr.components.Component | None]] | None
ChatbotTuplesOutput = List[List[str | Tuple[str] | Tuple[str, str] | None] | Tuple] | None

MultimodalTextboxInput = MultimodalValue | None
MultimodalTextboxOutput = MultimodalValue | None

TextboxInput = str | None
TextboxOutput = str | None

AudioInput = str | Tuple[int, np.ndarray] | None
AudioOutput = str | pathlib.Path | bytes | Tuple[int, np.ndarray] | None




import gradio as gr
import logging
import shirley as sh
import sys
from .interface import Interface


logger = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


class TextToSpeech(Interface):

    def __init__(self, local: bool = False, *args, **kwargs) -> None:
        super().__init__()

        self._client = sh.clients.TextToSpeech(local=local, *args, **kwargs)
        self._text: str = ''
        self._locale: str | None = None
        self._voice: str | None = None

        self._make_components(*args, **kwargs)


    @property
    def client(self) -> sh.clients.Chat:
        return self._client


    def _preconvert(self, *args, **kwargs) -> sh.types.GradioComponents:
        components = [
            gr.Textbox(interactive=False),
            gr.Button(variant='secondary', interactive=False),
            gr.ClearButton(interactive=False),
        ]
        return components


    def _convert(self, *args, **kwargs) -> sh.types.AudioOutput:
        return self._client.text_to_speech(text=self._text, voice=self._voice)


    def _postconvert(self, *args, **kwargs) -> sh.types.GradioComponents:
        components = [
            gr.Textbox(interactive=True),
            gr.Button(variant='primary', interactive=True),
            gr.ClearButton(interactive=True),
        ]
        return components


    def _submit(self, *args, **kwargs) -> None:
        textbox: sh.types.TextboxInput = args[0]

        if not textbox or not textbox.strip():
            raise gr.Error('Text not valid.')

        self._text = textbox


    def _reset(self, *args, **kwargs) -> sh.types.GradioComponents:
        self._text = ''

        return gr.Button(interactive=False)


    def _textbox_change(self, *args, **kwargs) -> sh.types.GradioComponents:
        textbox: sh.types.TextboxInput = args[0]

        if not textbox or not textbox.strip():
            return gr.Button(variant='secondary', interactive=False)

        self._text = textbox
        return gr.Button(variant='primary', interactive=True)


    def _locale_dropdown_change(self, *args, **kwargs) -> sh.types.GradioComponents:
        locale_dropdown: sh.types.DropdownInput = args[0]

        if not locale_dropdown or not locale_dropdown.strip():
            return gr.Dropdown(choices=None, interactive=False)

        self._locale = locale_dropdown
        voices = self._client.get_available_voices(locale=locale_dropdown)
        self._voice = None if not voices else voices[0]
        return gr.Dropdown(choices=voices, value=self._voice, interactive=True)


    def _voice_dropdown_change(self, *args, **kwargs) -> None:
        voice_dropdown: sh.types.DropdownInput = args[0]

        self._voice = voice_dropdown


    def _setup_textbox(self, *args, **kwargs) -> None:
        textbox: gr.Textbox = kwargs['textbox']
        convert_button: gr.Button = kwargs['convert_button']

        textbox.change(
            fn=self._textbox_change,
            inputs=[textbox],
            outputs=[convert_button],
            show_api=False,
        )


    def _setup_locale_dropdown(self, *args, **kwargs) -> None:
        locale_dropdown: gr.Dropdown = kwargs['locale_dropdown']
        voice_dropdown: gr.Dropdown = kwargs['voice_dropdown']

        locale_dropdown.change(
            fn=self._locale_dropdown_change,
            inputs=[locale_dropdown],
            outputs=[voice_dropdown],
            show_api=False,
        )


    def _setup_voice_dropdown(self, *args, **kwargs) -> None:
        voice_dropdown: gr.Dropdown = kwargs['voice_dropdown']

        voice_dropdown.change(
            fn=self._voice_dropdown_change,
            inputs=[voice_dropdown],
            outputs=None,
            show_api=False,
        )


    def _setup_convert_button(self, *args, **kwargs) -> None:
        textbox: gr.Textbox = kwargs['textbox']
        convert_button: gr.Button = kwargs['convert_button']
        reset_button: gr.ClearButton = kwargs['reset_button']
        audio: gr.Audio = kwargs['audio']

        submit = convert_button.click(
            fn=self._submit,
            inputs=[textbox],
            outputs=None,
            show_api=False,
        )
        preconvert = submit.success(
            fn=self._preconvert,
            inputs=None,
            outputs=[textbox, convert_button, reset_button],
            show_api=False,
        )
        convert = preconvert.then(
            fn=self._convert,
            inputs=None,
            outputs=[audio],
            show_api=False,
        )
        convert.then(
            fn=self._postconvert,
            inputs=None,
            outputs=[textbox, convert_button, reset_button],
            show_api=False,
        )


    def _setup_reset_button(self, *args, **kwargs) -> None:
        textbox: gr.Textbox = kwargs['textbox']
        convert_button: gr.Button = kwargs['convert_button']
        reset_button: gr.ClearButton = kwargs['reset_button']
        audio: gr.Audio = kwargs['audio']

        reset_button.add(components=[textbox, audio])

        reset_button.click(
            fn=self._reset,
            inputs=None,
            outputs=[convert_button],
            show_api=False,
        )


    def _setup(self, *args, **kwargs) -> None:
        self._setup_textbox(*args, **kwargs)
        self._setup_locale_dropdown(*args, **kwargs)
        self._setup_voice_dropdown(*args, **kwargs)
        self._setup_convert_button(*args, **kwargs)
        self._setup_reset_button(*args, **kwargs)


    def _make_components(self, *args, **kwargs) -> None:
        locales = self._client.get_available_locales()
        self._locale = locales[0]
        voices = self._client.get_available_voices(locale=self._locale)
        self._voice = None if not voices else voices[0]

        with gr.Row():
            with gr.Column():
                textbox = gr.Textbox(
                    lines=16,
                    max_lines=16,
                    placeholder='✏️ Enter text… (输入文字…)',
                    show_label=False,
                    show_copy_button=True,
                )
                with gr.Row():
                    locale_dropdown = gr.Dropdown(
                        choices=locales,
                        value=self._locale,
                        multiselect=False,
                        allow_custom_value=False,
                        label='🌏 Locale (语言)',
                    )
                    voice_dropdown = gr.Dropdown(
                        choices=voices,
                        value=self._voice,
                        multiselect=False,
                        allow_custom_value=False,
                        label='🎤 Voice (声音)',
                    )
                with gr.Row():
                    convert_button = gr.Button(value='🔄 Convert (转换)', variant='secondary', interactive=False)
                    reset_button = gr.ClearButton(value='🧹 Reset (重置)', variant='secondary')

            with gr.Column():
                audio = gr.Audio(
                    type='filepath',
                    show_label=False,
                    scale=1,
                    interactive=False,
                    show_download_button=True,
                )

        self._setup(
            textbox=textbox,
            locale_dropdown=locale_dropdown,
            voice_dropdown=voice_dropdown,
            convert_button=convert_button,
            reset_button=reset_button,
            audio=audio,
        )



import azure.cognitiveservices.speech as speechsdk
import logging
import os
import pathlib
import sys
import uuid
from .client import Client
from typing import List


logger = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


class TextToSpeech(Client):

    def __init__(self, local: bool, *args, **kwargs) -> None:
        super().__init__(local=local)

        self._speech_key: str | None = os.environ.get('SPEECH_KEY')
        self._speech_region: str | None = os.environ.get('SPEECH_REGION')


    def get_available_locales(self) -> List[str]:
        locales = [
            'zh-CN',
            'zh-CN-henan',
            'zh-CN-liaoning',
            'zh-CN-shaanxi',
            'zh-CN-shandong',
            'zh-CN-sichuan',
            'zh-HK',
            'zh-TW',
            'wuu-CN',
            'yue-CN',
            'en-US',
            'en-AU',
        ]
        return locales


    def get_available_voices(self, locale: str) -> List[str] | None:
        if not self._speech_key or not self._speech_region:
            return None

        speech_config = speechsdk.SpeechConfig(subscription=self._speech_key, region=self._speech_region)
        speech_synthesizer = speechsdk.SpeechSynthesizer(speech_config=speech_config, audio_config=None)

        result: speechsdk.SynthesisVoicesResult = speech_synthesizer.get_voices_async(locale).get()
        if result.reason == speechsdk.ResultReason.VoicesListRetrieved:
            logger.info('Voices successfully retrieved')
            return [voice.short_name for voice in result.voices]
        elif result.reason == speechsdk.ResultReason.Canceled:
            logger.error(f'Speech synthesis canceled; error details: {result.error_details}')


    def text_to_speech(self, text: str, voice: str) -> pathlib.Path | None:
        if not self._speech_key or not self._speech_region:
            return None

        speech_config = speechsdk.SpeechConfig(subscription=self._speech_key, region=self._speech_region)
        speech_config.speech_synthesis_voice_name = voice
        speech_config.set_speech_synthesis_output_format(
            format_id=speechsdk.SpeechSynthesisOutputFormat.Riff24Khz16BitMonoPcm,
        )

        audios_tempdir = pathlib.Path(self.tempdir) / 'audios'
        audios_tempdir.mkdir(exist_ok=True, parents=True)
        name = f'audio-{uuid.uuid4()}.wav'
        filename = audios_tempdir / name
        audio_config = speechsdk.audio.AudioOutputConfig(filename=str(filename))

        speech_synthesizer = speechsdk.SpeechSynthesizer(
            speech_config=speech_config,
            audio_config=audio_config,
        )

        speech_synthesis_result: speechsdk.SpeechSynthesisResult = speech_synthesizer.speak_text_async(text).get()

        if speech_synthesis_result.reason == speechsdk.ResultReason.SynthesizingAudioCompleted:
            logger.info(f'Speech synthesized for text [{text}]')
            logger.info(f'Audio file saved in {str(filename)}.')
            return filename
        elif speech_synthesis_result.reason == speechsdk.ResultReason.Canceled:
            cancellation_details = speech_synthesis_result.cancellation_details
            logger.error(f'Speech synthesis canceled: {cancellation_details.reason}')
            if cancellation_details.reason == speechsdk.CancellationReason.Error:
                if cancellation_details.error_details:
                    logger.error(f'Error details: {cancellation_details.error_details}')
                    logger.error('Did you set the speech resource key and region values?')
            return None



import gradio as gr
import logging
import pypdf
import shirley as sh
import sys
from .interface import Interface
from gradio.events import Dependency
from typing import Callable, Iterator, List, Tuple


logger = logging.getLogger(__name__)
logging.basicConfig(stream=sys.stdout, level=logging.INFO)


class Chat(Interface):

    def __init__(self, local: bool = True, *args, **kwargs) -> None:
        super().__init__()

        self._client: sh.clients.Chat = sh.clients.Chat(local=local, *args, **kwargs)
        self._chat_stream: Callable = kwargs.get('chat_stream', self._client.chat_stream)
        self._draw_bbox_on_latest_picture = self._client.draw_bbox_on_latest_picture
        self._pretrained_models: List[str] = self._client.get_models()
        self._pretrained_model_name_or_path: str = self._client.get_model_name_or_path(
            model_name=self._pretrained_models[0],
        )
        self._generating: bool = False
        self._history: List[Tuple] = []

        self._make_components(*args, **kwargs)


    def _load_context(self, filepath: str) -> str:
        if sh.utils.isimage(filepath):
            return f'Picture: <img>{filepath}</img>'
        elif filepath.endswith('.pdf'):
            reader = pypdf.PdfReader(stream=filepath)
            return '\n'.join([page.extract_text() for page in reader.pages])
        else:
            logger.warning(f'File type {filepath} not supported.')
            return ''


    def _get_query_and_history(self) -> Tuple[sh.types.QwenQuery, sh.types.QwenHistory]:
        history: sh.types.QwenHistory = []
        text = ''
        for _, (query, response) in enumerate(self._history):
            if isinstance(query, (Tuple, List)):
                filepath = query[0]
                context = self._load_context(filepath=filepath)
                text += context + '\n'
            else:
                text += query
                history.append((text, response))
                text = ''
        return history[-1][0], history[:-1]


    def _pregenerate(self, *args, **kwargs) -> sh.types.GradioComponents:
        components = [
            gr.MultimodalTextbox(interactive=False),
            gr.Button(variant='secondary', interactive=False),
            gr.Button(variant='stop', interactive=True),
            gr.Button(interactive=False),
            gr.ClearButton(interactive=False),
        ]
        return tuple(components)


    def _generate(self, *args, **kwargs) -> Iterator[sh.types.ChatbotTuplesOutput]:
        chatbot: sh.types.ChatbotTuplesInput = args[0]

        self._generating = True
        logger.info(f'😀 User: {chatbot[-1][0]}')

        query, history = self._get_query_and_history()
        full_response = ''
        for response in self._chat_stream(query=query, history=history):
            if not self._generating: break
            chatbot[-1] = (chatbot[-1][0], sh.utils.parse(text=response, remove_image_tags=True))
            yield chatbot
            full_response = sh.utils.parse(text=response)

        history.append((query, full_response))
        image_filepath = self._draw_bbox_on_latest_picture(history=history)
        if image_filepath is not None:
            chatbot.append((None, (image_filepath,)))
        else:
            chatbot[-1] = (chatbot[-1][0], full_response)
        self._history[-1] = (self._history[-1][0], full_response)

        logger.info(f'🦈 Shirley: {full_response}')
        self._generating = False
        yield chatbot


    def _postgenerate(self, *args, **kwargs) -> sh.types.GradioComponents:
        components = [
            gr.MultimodalTextbox(interactive=True),
            gr.Button(variant='secondary', interactive=False),
            gr.Button(variant='secondary', interactive=False),
            gr.Button(interactive=True),
            gr.ClearButton(interactive=True),
        ]
        return tuple(components)


    def _validate(self, *args, **kwargs) -> None:
        if not self._client.model:
            logger.error('Model not loaded.')
            raise gr.Error('Model not loaded. Please load a model.')


    def _submit(self, *args, **kwargs) -> Tuple[sh.types.ChatbotTuplesOutput, sh.types.MultimodalTextboxOutput]:
        chatbot: sh.types.ChatbotTuplesInput = args[0]
        multimodal_textbox: sh.types.MultimodalTextboxInput = args[1]

        text = multimodal_textbox['text']
        if not text or not text.strip():
            raise gr.Error('Text not valid.')

        for filepath in multimodal_textbox['files']:
            chatbot = chatbot + [((filepath,), None)]
            self._history = self._history + [((filepath,), None)]

        if multimodal_textbox['text'] is not None:
            chatbot = chatbot + [(sh.utils.parse(text=multimodal_textbox['text']), None)]
            self._history = self._history + [(multimodal_textbox['text'], None)]

        return chatbot, None


    def _stop(self, *args, **kwargs) -> sh.types.GradioComponents:
        self._generating = False


    def _regenerate(self, *args, **kwargs) -> sh.types.ChatbotTuplesOutput | Iterator[sh.types.ChatbotTuplesOutput]:
        chatbot: sh.types.ChatbotTuplesInput = args[0]

        if len(chatbot) < 1 or len(self._history) < 1:
            return chatbot

        history_last = self._history[-1]
        if history_last[1] is None:
            return chatbot
        self._history[-1] = (history_last[0], None)

        chatbot_last = chatbot.pop(-1)
        if chatbot_last[0] is None:
            chatbot[-1] = (chatbot[-1][0], None)
        else:
            chatbot.append((chatbot_last[0], None))

        yield from self._generate(*args, **kwargs)


    def _reset(self, *args, **kwargs) -> sh.types.GradioComponents:
        self._history = []

        components = [
            gr.MultimodalTextbox(interactive=True),
            gr.Button(variant='secondary', interactive=False),
            gr.Button(variant='secondary', interactive=False),
            gr.Button(interactive=False),
            gr.ClearButton(interactive=False),
        ]
        return tuple(components)


    def _model_dropdown_change(self, *args, **kwargs) -> None:
        model_dropdown: sh.types.DropdownInput = args[0]

        self._pretrained_model_name_or_path = self._client.get_model_name_or_path(model_name=model_dropdown)


    def _load_button_click(self, *args, **kwargs) -> sh.types.GradioComponents:
        self._client.load_model(pretrained_model_name_or_path=self._pretrained_model_name_or_path)
        gr.Info(message='Model loaded.')

        return gr.Dropdown(interactive=True), gr.Button(interactive=True)


    def _multimodal_textbox_change(self, *args, **kwargs) -> sh.types.GradioComponents:
        multimodal_textbox: sh.types.MultimodalTextboxInput = args[0]

        text = multimodal_textbox['text']
        if not text or not text.strip():
            return gr.Button(variant='secondary', interactive=False)

        return gr.Button(variant='primary', interactive=True)


    def _setup_model_dropdown(self, *args, **kwargs) -> None:
        model_dropdown: gr.Dropdown = kwargs['model_dropdown']

        model_dropdown.change(
            fn=self._model_dropdown_change,
            inputs=[model_dropdown],
            outputs=None,
            show_api=False,
        )


    def _setup_load_button(self, *args, **kwargs) -> None:
        model_dropdown: gr.Dropdown = kwargs['model_dropdown']
        load_button: gr.Button = kwargs['load_button']
        model_config: gr.JSON = kwargs['model_config']

        preload = load_button.click(
            fn=lambda: (gr.Dropdown(interactive=False), gr.Button(interactive=False)),
            inputs=None,
            outputs=[model_dropdown, load_button],
            show_api=False,
        )
        load_button_click = preload.then(
            fn=self._load_button_click,
            inputs=None,
            outputs=[model_dropdown, load_button],
            show_api=False,
        )
        load_button_click.then(
            fn=lambda: self._client.get_model_config(),
            inputs=None,
            outputs=[model_config],
            show_api=False,
        )


    def _set_event_trigger_generate(self, dependency: Dependency, fn: Callable, *args, **kwargs) -> None:
        chatbot: gr.Chatbot = kwargs['chatbot']
        multimodal_textbox: gr.MultimodalTextbox = kwargs['multimodal_textbox']
        submit_button: gr.Button = kwargs['submit_button']
        stop_button: gr.Button = kwargs['stop_button']
        regenerate_button: gr.Button = kwargs['regenerate_button']
        reset_button: gr.ClearButton = kwargs['reset_button']

        success = dependency.success(fn=lambda:None, show_api=False)
        pregenerate = success.then(
            fn=self._pregenerate,
            inputs=None,
            outputs=[multimodal_textbox, submit_button, stop_button, regenerate_button, reset_button],
            show_api=False,
        )
        generate = pregenerate.then(
            fn=fn,
            inputs=[chatbot],
            outputs=[chatbot],
            show_api=False,
        )
        generate.then(
            fn=self._postgenerate,
            inputs=None,
            outputs=[multimodal_textbox, submit_button, stop_button, regenerate_button, reset_button],
            show_api=False,
        )


    def _setup_multimodal_textbox(self, *args, **kwargs) -> None:
        chatbot: gr.Chatbot = kwargs['chatbot']
        multimodal_textbox: gr.MultimodalTextbox = kwargs['multimodal_textbox']
        submit_button: gr.Button = kwargs['submit_button']

        multimodal_textbox.change(
            fn=self._multimodal_textbox_change,
            inputs=[multimodal_textbox],
            outputs=[submit_button],
            show_api=False,
        )

        validate = multimodal_textbox.submit(
            fn=self._validate,
            show_api=False,
        )
        submit = validate.success(
            fn=self._submit,
            inputs=[chatbot, multimodal_textbox],
            outputs=[chatbot, multimodal_textbox],
            show_api=False,
        )
        self._set_event_trigger_generate(dependency=submit, fn=self._generate, *args, **kwargs)


    def _setup_submit_button(self, *args, **kwargs) -> None:
        chatbot: gr.Chatbot = kwargs['chatbot']
        multimodal_textbox: gr.MultimodalTextbox = kwargs['multimodal_textbox']
        submit_button: gr.Button = kwargs['submit_button']

        validate = submit_button.click(
            fn=self._validate,
            show_api=False,
        )
        submit = validate.success(
            fn=self._submit,
            inputs=[chatbot, multimodal_textbox],
            outputs=[chatbot, multimodal_textbox],
            show_api=False,
        )
        self._set_event_trigger_generate(dependency=submit, fn=self._generate, *args, **kwargs)


    def _setup_stop_button(self, *args, **kwargs) -> None:
        stop_button: gr.Button = kwargs['stop_button']

        stop_button.click(fn=self._stop, show_api=False)


    def _setup_regenerate_button(self, *args, **kwargs) -> None:
        regenerate_button: gr.Button = kwargs['regenerate_button']

        click = regenerate_button.click(fn=lambda:None, show_api=False)
        self._set_event_trigger_generate(dependency=click, fn=self._regenerate, *args, **kwargs)


    def _setup_reset_button(self, *args, **kwargs) -> None:
        chatbot: gr.Chatbot = kwargs['chatbot']
        multimodal_textbox: gr.MultimodalTextbox = kwargs['multimodal_textbox']
        submit_button: gr.Button = kwargs['submit_button']
        stop_button: gr.Button = kwargs['stop_button']
        regenerate_button: gr.Button = kwargs['regenerate_button']
        reset_button: gr.ClearButton = kwargs['reset_button']

        reset_button.add(components=[chatbot, multimodal_textbox])

        reset_button.click(
            fn=self._reset,
            inputs=None,
            outputs=[multimodal_textbox, submit_button, stop_button, regenerate_button, reset_button],
            show_api=False,
        )


    def _setup(self, *args, **kwargs) -> None:
        self._setup_model_dropdown(*args, **kwargs)
        self._setup_load_button(*args, **kwargs)
        self._setup_multimodal_textbox(*args, **kwargs)
        self._setup_submit_button(*args, **kwargs)
        self._setup_stop_button(*args, **kwargs)
        self._setup_regenerate_button(*args, **kwargs)
        self._setup_reset_button(*args, **kwargs)


    def _make_components(self, *args, **kwargs) -> None:
        avatar_images: Tuple | None = kwargs.get('avatar_images', None)

        with gr.Row():
            with gr.Column(scale=1):
                with gr.Group():
                    model_dropdown = gr.Dropdown(
                        choices=self._pretrained_models,
                        value=self._pretrained_models[0],
                        multiselect=False,
                        allow_custom_value=False,
                        label='🤗 Model (模型)',
                    )
                    load_button = gr.Button(value='📥 Load (加载)', variant='secondary')
                model_config = gr.JSON(show_label=False, scale=1)

            with gr.Column(scale=3):
                chatbot = gr.Chatbot(
                    show_label=False,
                    height='50vh',
                    show_copy_button=True,
                    avatar_images=avatar_images,
                )
                multimodal_textbox = gr.MultimodalTextbox(
                    placeholder='✏️ Enter text or upload file… (输入文字或者上传文件…)',
                    show_label=False,
                    interactive=True,
                )
                with gr.Row():
                    submit_button = gr.Button(value='🚀 Submit (发送)', variant='secondary', interactive=False)
                    stop_button = gr.Button(value='⏹️ Stop (停止)', variant='secondary', interactive=False)
                    regenerate_button = gr.Button(value='🔁 Regenerate (重新生成)', interactive=False)
                    reset_button = gr.ClearButton(value='🧹 Reset (重置)', interactive=False)

        self._setup(
            model_dropdown=model_dropdown,
            load_button=load_button,
            model_config=model_config,
            chatbot=chatbot,
            multimodal_textbox=multimodal_textbox,
            submit_button=submit_button,
            stop_button=stop_button,
            regenerate_button=regenerate_button,
            reset_button=reset_button,
        )
